from .csp import *
from .enums import *
from .exceptions import *
from .fastapi import query_json
from .id_generator import get_counter
from .picklable_queue import PickleableQueue
from .struct import GatewayStruct, IdType
from .web import *



from .client import (
    AsyncGatewayClient,
    AsyncGatewayClientMixin,
    BaseGatewayClient,
    GatewayClient,
    GatewayClientConfig,
    ResponseWrapper,
    SyncGatewayClient,
    SyncGatewayClientMixin,
)



from datetime import datetime, timezone
from types import MappingProxyType
from typing import Any, Dict

import csp
from csp import Struct
from csp.impl.struct import StructMeta
from pydantic import ValidationInfo
from pydantic_core import CoreConfig, core_schema

from ..id_generator import get_counter
from .psp import PerspectiveUtilityMixin

IdType = str

__all__ = ("GatewayStruct", "IdType", "PydanticizedCspStruct")


class PydanticizedCspStruct(StructMeta):
    """A subclass of StructMeta from csp, this class adds additional properties onto csp.Struct classes to link them into the Gateway format.

    Specifically, allows lookups and automatic id generation.
    """

    def __init__(cls: Any, name: str, bases: Any, attr_dict: Any) -> None:
        super().__init__(name, bases, attr_dict)
        # Automatically construct pydantic model from csp.struct

        # Attach an id generator to every class
        cls.id_generator = get_counter(cls)

        # Allow for looking up by ID
        cls._internal_mapping: Dict[str, Any] = {}

        # But expose this lookup as a readonly mapping proxy
        cls.lookup = MappingProxyType(cls._internal_mapping).get
        cls._include_in_lookup = True

    def omit_from_lookup(cls, omit=True):
        cls._include_in_lookup = not omit

    def included_in_lookup(cls):
        return cls._include_in_lookup


class GatewayStruct(PerspectiveUtilityMixin, Struct, metaclass=PydanticizedCspStruct):
    """Sub-class of csp.Struct specifically designed for usage with csp-gateway.
    These classes inherit from csp.Struct, but each one also contains a pydantic model
    as an attribute that mirrors the underlying struct class.

    The pydantic model can be constructed by running `.to_pydantic()` on an instance of the
    given GatewayStruct. To access the underlying class, one can call `__pydantic_model__` on the specific GatewayStruct sub-class object.
    For example:
        class MyClass(GatewayStruct):
            x: int
        pydantic_model = MyClass.__pydantic_model__

    The pydantic model of the the GatewayStruct allows for data validation. Furthermore, the utilities
    in csp-gateway (such as Rest routes, Kafka, Perspective integrations) are specifically designed for GatewayStruct's.
    """

    id: IdType
    timestamp: datetime

    def __init__(self, **kwargs: Any) -> None:
        # auto generate id on every new construction
        if "id" not in kwargs:
            # TODO consider postfixing with a _ for conflicts
            kwargs["id"] = str(self.__class__.id_generator.next())
        if "timestamp" not in kwargs:
            kwargs["timestamp"] = datetime.now(timezone.utc)

        # And put into lookup
        # TODO make this nicer when we switch to pydantic as first class instead of csp struct
        if self.__class__.included_in_lookup():
            self.__class__._internal_mapping[kwargs["id"]] = self

        # and defer to normal csp.struct construction
        super().__init__(**kwargs)

    @classmethod
    def _validate_gateway_struct_after(cls, val):
        """Validate GatewayStruct after pydantic type validation.
        A validator attached to every GatewayStruct to allow for defining custom
        model-level after validators that run after pydantic type validation.
        If not defined on a child class, the parent's validator will be used.  If defined on a child class, the parent's validator will be ignored. Please call the parent's validator directly if you want to run both.
        Args:
            cls: The class this validator is attached to
            val: The value to validate
        Returns:
            The validated value, possibly modified
        """
        return val

    @classmethod
    def type_adapter(cls):
        # NOTE: Only needed until csp>0.9 is released with this fix
        # We mangle ourselves, explicitly, to make sure that child Structs
        # will get their own type adapters.
        attr_name = f"_{cls.__name__}__pydantic_type_adapter"
        internal_type_adapter = getattr(cls, attr_name, None)
        if internal_type_adapter:
            return internal_type_adapter

        # Late import to avoid autogen issues
        from pydantic import TypeAdapter

        type_adapter = TypeAdapter(cls)
        setattr(cls, attr_name, type_adapter)
        return type_adapter

    @classmethod
    def _validate_gateway_struct(cls, val, handler, info: ValidationInfo):
        if isinstance(info.context, dict) and isinstance(val, dict):
            if info.context.get("force_new_id", False):
                # If we are forcing a new id, we need to remove the old one
                val.pop("id", None)
            if info.context.get("force_new_timestamp", False):
                # If we are forcing a new timestamp, we need to remove the old one
                val.pop("timestamp", None)
        csp_struct = handler(val)
        final = cls._validate_gateway_struct_after(csp_struct)
        return final

    @staticmethod
    def _get_pydantic_core_schema(cls, source_type, handler):
        # Get parent schema - note the cls parameter
        parent_schema = csp.Struct._get_pydantic_core_schema(cls, source_type, handler)
        core_config = CoreConfig(coerce_numbers_to_str=True)
        # soooo hacky...
        parent_schema["schema"]["config"] = core_config
        return core_schema.with_info_wrap_validator_function(
            function=cls._validate_gateway_struct, schema=parent_schema, serialization=parent_schema.get("serialization")
        )

    @classmethod
    def generate_id(cls) -> str:
        return str(cls.id_generator.next())



import asyncio
import logging

from fastapi import APIRouter, BackgroundTasks, HTTPException, Request

from csp_gateway.utils import Controls

from ..utils import get_default_responses
from .shared import prepare_response

log = logging.getLogger(__name__)


_WAIT_THRESHOLD = 0.1


def add_controls_routes(api_router: APIRouter) -> None:
    # Add heartbeat channel
    @api_router.get(
        "/heartbeat",
        responses=get_default_responses(),
        response_model=Controls,
        name="Get Heartbeat",
    )
    async def heartbeat(request: Request) -> Controls:
        """
        This endpoint is a lightweight `ping`/`pong` endpoint that can be used to determine the status of the underlying webserver.
        """
        data = Controls(name="heartbeat")

        # Throw 404 if not a supported channel
        if not hasattr(request.app.gateway.channels, "controls"):
            raise HTTPException(status_code=404, detail="Channel not found: controls")

        # send data to csp
        request.app.gateway.channels.send("controls", data)

        # don't care about the result
        while data.status != "ok":
            await asyncio.sleep(_WAIT_THRESHOLD)

        return prepare_response(data, is_list_model=False)

    @api_router.get(
        "/stats",
        responses=get_default_responses(),
        response_model=Controls,
        name="Get CSP Stats",
    )
    async def stats(request: Request) -> Controls:
        """This endpoint will collect and return various engine and system stats, including:

        - CPU utilization (`cpu`)
        - Virtual memory utilization (`memory`)
        - Total memory available (`memory-total`)
        - Current system time (`now`)
        - CSP engine time (`csp-now`)
        - Hostname (`host`)
        - Username (`user`)
        """
        data = Controls(name="stats")

        # Throw 404 if not a supported channel
        if not hasattr(request.app.gateway.channels, "controls"):
            raise HTTPException(status_code=404, detail="Channel not found: controls")

        # send data to csp
        request.app.gateway.channels.send("controls", data)

        while not data.data:
            await asyncio.sleep(_WAIT_THRESHOLD)
        data.update_str()

        return prepare_response(data, is_list_model=False)

    @api_router.post(
        "/shutdown",
        responses=get_default_responses(),
        response_model=Controls,
        name="Shutdown Server",
    )
    async def shutdown(request: Request, background_tasks: BackgroundTasks) -> Controls:
        """
        **WARNING:** Use this endpoint with caution.

        This endpoint will cleanly shutdown the engine and webserver. It is used for the kill switch in UIs.
        """
        # FIXME ugly
        background_tasks.add_task(request.app.gateway.stop, user_initiated=True)

        data = Controls(name="shutdown", status="ok")
        return prepare_response(data, is_list_model=False)



from .controls import add_controls_routes
from .last import add_last_api_available_channels, add_last_routes
from .lookup import add_lookup_routes
from .next import add_next_api_available_channels, add_next_routes
from .send import add_send_api_available_channels, add_send_routes
from .shared import prepare_response
from .state import add_state_api_available_channels, add_state_routes



from csp_gateway import GatewayStruct as Base


class LookupModel(Base):
    foo: int = 9


class NoLookupModel(Base):
    foo: int = 10


NoLookupModel.omit_from_lookup(True)


def test_automatic_id_generation():
    for Model in [LookupModel, NoLookupModel]:
        o1 = Model()
        value1 = str(Model.id_generator.current())
        assert o1.id == value1

        o2 = Model()
        value2 = str(Model.id_generator.current())
        assert o2.id == value2
        assert o2.id == str(int(o1.id) + 1)

        if Model == LookupModel:
            assert Model.lookup(value1) == o1
            assert Model.lookup(value2) == o2


def test_lookup_fails():
    o1 = LookupModel()
    value1 = str(LookupModel.id_generator.current())
    assert o1.id == value1

    o2 = LookupModel()
    value2 = str(LookupModel.id_generator.current())
    assert o2.id == value2
    assert o2.id == str(int(o1.id) + 1)

    assert LookupModel.lookup(value1) == o1
    assert LookupModel.lookup(value2) == o2

    o1 = NoLookupModel()
    value1 = str(NoLookupModel.id_generator.current())
    assert o1.id == value1

    o2 = NoLookupModel()
    value2 = str(NoLookupModel.id_generator.current())
    assert o2.id == value2
    assert o2.id == str(int(o1.id) + 1)

    assert NoLookupModel.lookup(value1) is None
    assert NoLookupModel.lookup(value2) is None



import logging
from datetime import datetime
from enum import Enum as PyEnum
from typing import Literal, Optional, Union

from csp.impl.enum import Enum as CspEnum, EnumMeta as CspEnumMeta
from pydantic import BaseModel, Field

log = logging.getLogger(__name__)

FilterWhere = Literal["==", "!=", "<", "<=", ">", ">="]
FilterWhereLambdaMap = {
    "==": lambda a, b: a == b,
    "!=": lambda a, b: a != b,
    "<": lambda a, b: a < b,
    "<=": lambda a, b: a <= b,
    ">": lambda a, b: a > b,
    ">=": lambda a, b: a >= b,
}


class FilterCondition(BaseModel):
    # in priority order
    value: Optional[Union[float, int, str]] = Field(None)
    # have to handle separately otherwise
    # all ints would be datetimes..
    when: Optional[datetime] = Field(None)
    attr: str = ""
    where: FilterWhere = "=="


class Filter(BaseModel):
    attr: str = ""
    by: FilterCondition

    def calculate(self, obj) -> bool:
        """
        calculate the filter condition on the object

        returns `True` if SHOULD NOT BE FILTERED, else `False`
        """
        try:
            if self.by.value is not None:
                lhs = getattr(obj, self.attr)
                # Convert enums attrs to strings during filtering
                if isinstance(lhs, (CspEnum, PyEnum, CspEnumMeta)):
                    lhs = lhs.name
                log.info(f"Filtering: {lhs} {self.by.where} {self.by.value}")
                return FilterWhereLambdaMap[self.by.where](lhs, self.by.value)
            if self.by.when is not None:
                log.info(f"Filtering: {getattr(obj, self.attr)} {self.by.where} {self.by.when}")
                return FilterWhereLambdaMap[self.by.where](getattr(obj, self.attr), self.by.when)
            if self.by.attr:
                log.info(f"Filtering: {getattr(obj, self.attr)} {self.by.where} {getattr(obj, self.by.attr)}")
                return FilterWhereLambdaMap[self.by.where](getattr(obj, self.attr), getattr(obj, self.by.attr))
        except (ValueError, AttributeError) as e:
            # TODO probably surface to webserver
            log.warning(f"Error during filtering {type(obj)} / {self}: {e}")

        # default case, if there was an issue assume its not included
        return False
